
/**
 * Preliminary start to this.
 * 
 * @author Peter Dordal 
 * @version 0.01
 */
public class SymbolTable
{
    // local vars: markers for next global and next local variable, 
    // two Map<String, IdentInfo> objects for global and local vars, respectively
    // some indication of whether the current scope is global or local, 
    // eg a boolean isGlobal, or just whether or not the local map is null.
    
    Tokenizer theTokenizer;
    
    /**
     * Constructor for objects of class SymbolTable
     * Tokenizer added as parameter so we can print out the line 
     * if a duplicate declaration is seen.
     */
    public SymbolTable(Tokenizer t)
    {
        // initialise instance variables
        
        theTokenizer = t;
    }

    /**
     * allocate() creates an entry in the CURRENT scope for the given identifier.
     * The current scope is local if isGlobal is false, otherwise global.
     * 
     * It is an error if there is already an existing entry for that identifier
     * in that scope; this can be indicated by having allocate() return null
     * (paradoxical for a not-found case!) Entries in previous scopes do not matter. 
     * If the identifier is found, of course, then its IdentInfo is returned.
     */
    
    private IdentInfo allocate(String ident, IdentInfo ii) {
        return ii;
    }
    
    // This allocates a procedure and fills in the data
    public IdentInfo allocProc(String fname, int offset) {
        IdentInfo theRec = new IdentInfo();
        theRec.setType(IdentInfo.IDType.PROCEDURE);
        theRec.setEntryPoint(offset);
        
        return allocate(fname, theRec);
    }
    
    public IdentInfo allocVar(String vname, boolean isGlobal) {
        IdentInfo theRec = new IdentInfo();
        theRec.setType(IdentInfo.IDType.VARIABLE);
        theRec.setIsGlobal(isGlobal);
        
        int offset = 0;         // pld12: you need to have allocVar() set this 
                                // to the next available global or local offset
        theRec.setAddr(offset);

        return allocate(vname, theRec);
        }
    
    public IdentInfo allocConst(String cname, int value) {
        IdentInfo theRec = new IdentInfo();
        theRec.setType(IdentInfo.IDType.CONSTANT);
        theRec.setValue(value);
        // lookup and install theRec
        return allocate(cname, theRec);
    }

    
    /**
     * lookup() finds the IdentInfo for the given identifier, in the "most recent"
     * scope. It is a compiler error if the identifier is not found,
     * although it is sufficient for lookup() to return null in this case
     * and have, say, compileFactor() issue the actual error.
     * lookup() should *never* create new records.
     */
    
    public IdentInfo lookup(String ident) {
        return null;
    }
    /**
     * locSize: returns the number of LOCAL variables in the table
     */
    public int stackFrameSize() {
        // put your code here
        return 0;
    }
    
    /**
     * creates a new scope, "freezing" somehow the state of the existing table
     */
    public void newScope() {
    }
    
    /**
     * endScope() has to restore the table to its previous status.
     * Either here or at the end of compileCompound(), where this is called,
     * we also need cs.emit(Machine.ALLOC, stackFrameSize()),
     * where the size is the size *after* the end.
     */
    public void endScope() {
    }
    
    // use this for reporting duplicate declarations in allocate()
    private void error(String s) {theTokenizer.error(s);}

}
